# State Pool

![Build Status](https://github.com/yezyilomo/state-pool/actions/workflows/node.js.yml/badge.svg?branch=master)
[![Build Size](https://img.shields.io/bundlephobia/minzip/state-pool?label=bundle-size&style=flat)](https://bundlephobia.com/result?p=state-pool)
[![Version](https://img.shields.io/npm/v/state-pool?style=flat)](https://www.npmjs.com/package/state-pool)
[![Downloads](https://img.shields.io/npm/dt/state-pool.svg?style=flat)](https://www.npmjs.com/package/state-pool)

React state management library based on global variables and react hooks.

## Features & Advantages
- Simple, familiar, flexible and very minimal core API but powerful
- Built-in support for state persistence
- Very easy to learn because its API is very similar to react state hook's API
- Support selecting deeply nested state
- Support creating global state dynamically
- Can be used outside react components
- Support both key based and non-key based global state
- States are stored as global variables(Can be used anywhere)
- Doesn't wrap your app in context providers
- Very organized API, You can do almost everything with a single import

<br/>

## Architectural Diagram
![Architecture Diagram](https://raw.githubusercontent.com/yezyilomo/state-pool/master/docs/images/architecture_diagram.png)

<br/>

### State Flow
1. Create a global state

2. Subscribe a component(s) to a created global state

3. If a component wants to update a global state, it sends update request

4. When a global state receives update request, it performs the update and send signal to all components subscribed to it for them to update themselves(re-render)

<br/>

## Installing
```
yarn add state-pool
```

Or 

```
npm install state-pool
```

<br/>

## Getting Started
Using **state-pool** to manage global state is very simple, all you need to do is
1. Create a store(which is basically a container for your global state)
1. Create and initialize a global state by using `store.setState`
2. Use your global state in your component through `store.useState` hooks

These three steps summarises pretty much everything you need to use **state-pool**.

Below are few examples showing how to use **state-pool** to manage global states.

```js
// Example 1.
import React from 'react';
import { createStore } from 'state-pool';


const store = createStore();  // Create store for storing our global state
store.setState("count", 0);  // Create "count" global state and add it to the store

function ClicksCounter(props){
    // Use "count" global state
    const [count, setCount] = store.useState("count");

    const incrementCount = (e) => {
        setCount(count+1)
    }

    return (
        <div>
            Count: {count}
            <br/>
            <button onClick={incrementCount}>Click</button>
        </div>
    );
}

ReactDOM.render(ClicksCounter, document.querySelector("#root"));
```

<br/>

```js
// Example 2.
import React from 'react';
import { createStore } from 'state-pool';


const store = createStore();  // Create store for storing our global state
store.setState("user", {name: "Yezy", age: 25});

function UserInfo(props){
    const [user, setUser, updateUser] = store.useState("user");

    const updateName = (e) => {
        updateUser(user => {
            user.name = e.target.value;
        });
    }

    return (
        <div>
            Name: {user.name}
            <br/>
            <input type="text" value={user.name} onChange={updateName}/>
        </div>
    );
}

ReactDOM.render(UserInfo, document.querySelector("#root"));
```

<br/>

## API
### createStore
Store is a container for global states, it comes with several methods which are used to manage states in it including `store.setState`, `store.getState`, `store.remove`, `store.clear`, `store.useState`, `store.useReducer` and `store.subscribe`. Store is created by using `createStore` API as 

```js
import { createStore } from 'state-pool';

const store = createStore();
```

<br/>

### store.setState
This is used to create a global state and map it to a key so that you won't need to use it directly, instead you use a key to get it. `store.setState` takes two required parameters, a key(string) to map to a global state object and the initial value, In addition to those two parameters it takes a third optional parameter which is the configuration object. `persist` is the only available config which is the flag to determine whether to save/persist state in a permanent storage or not.

```js
// Signature
store.setState(key: String, initialState: Any, config?: {persist: Boolean})
```

Here is how to use it

```js
const userState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}
store.setState("user", userState);
```

**Note:** `store.setState` should be used outside of a react component.

<br/>

### store.useState hook
`store.useState` is a hook that used to get a global state from a store, it's a way for a component to subscribe to a global state from a store. `store.useState` works just like `React.useState` hook but it accepts a key for the global state and returns an array of `[state, setState, updateState]` rather than `[state, setState]`. In addition to the key parameter it also accept another optional parameter which is the config object, available configurations are `default`, `persist`, `selector` & `patcher`, these will be discussed in detail later.

```js
// Signature
store.useState(
    key: String,
    config?: {default: Any, persist: Boolean, selector: Function, patcher: Function}
)
```

Below is an example showing how to use `store.useState` hook

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

store.setState("user", initialGlobalState);

function Component(props){
    const [user, setUser, updateUser] = store.useState("user");
    // Other stuff
}
```

Here `updateUser` is a higher order function which accepts another function for updating user as an argument(this another functions takes user(old state) as the argument). So to update any value on user you could do

```js
updateUser(function(user){
    user.name = "Yezy Ilomo";
    user.age = 26;
})
```

You could also return a new state instead of changing it i.e

```js
updateUser(function(user){
    return {
        name: "Yezy Ilomo",
        age: 26,
        email: "yezy@me.com"
    }
})
```

Or you could just use `setUser` instead of `updateUser` i.e

```js
setUser({name: "Yezy Ilomo", age: 26, email: "yezy@me.com"});
```

As stated earlier `store.useState` takes a second optional parameter which is a configuration object, available configurations are:
- `default` - This is used to specify the default value if you want `store.useState` to create a global state if it doesn't find the one for the key specified in the first argument. For example

  ```js
  const [user, setUser, updateUser] = store.useState("user", {default: null});
  ```

  This piece of code means, get the global state for the key "user" if it's not available in a store, create one and assign it the value `null`.

- Also in addition to `default` configuration there is `persist` configuration which is the flag to determine whether to save/persist global state in your preferred storage or not. Here persist configuration is only used if `store.useState` is going to create global state dynamically.
<br/>

Other allowed configurations are `selector` & `patcher`. These are used for specifying a way to select deeply nested state and update it.

- `selector` should be a function which takes one parameter which is the global state and returns a selected value. The purpose of this is to subscribe to a deeply nested state.

- `patcher` should be a function which takes two parameters, the first is the global state and the second is the selected value. The purpose of this is to merge back the selected value to the global state once it's updated.

  Example.
  ```js
  const initialGlobalState = {
      name: "Yezy",
      age: 25,
      email: "yezy@me.com"
  }
  
  store.setState("user", initialGlobalState);
  
  
  function UserName(props){
      const selector = (user) => user.name;  // Subscribe to user.name only
      const patcher = (user, name) => {user.name = name};  // Update user.name
  
      const [name, setName] = store.useState("user", {selector: selector, patcher: patcher});
  
      let handleNameChange = (e) => {
          setName(e.target.value);
      }
  
      return (
          <div>
              Name: {name}
              <br/>
              <input type="text" value={name} onChange={handleNameChange}/>
          </div>
      );
  }
  ```

<br/>

### store.useReducer hook
This is an alternative to `store.useState`, it works just like `React.useReducer` hook((If you’re familiar with `React.useReducer`, you already know how this works). It accepts a reducer and a key for the global state as parameters, it returns the current state paired with a dispatch method. In addition to the two parameters it also accept another optinal perameter which is the configuration object, available configurations are `default`, `persist`, `selector` & `patcher` they work exactly the same just like in `store.useState`.

```js
// Signature
store.useReducer(
    reducer: Function,
    key: String,
    config?: {default: Any, persist: Boolean, selector: Function, patcher: Function}
)
```

Below is a simple example showing how to use `store.useReducer` hook

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

store.setState("user", initialGlobalState);

function myReducer(state, action){
    // This could be any reducer
    // Do whatever you want to do here
    return newState
}

function Component(props){
    const [name, dispatch] = store.useReducer(myReducer, "user");

    // Other stuff ...
}
```


Below is an example with `selector` and `patcher` parameters

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

store.setState("user", initialGlobalState);


function myReducer(state, action){
    // This could be any reducer
    // Do whatever you want to do here
    return newState
}

function UserInfo(props){
    const selector = (user) => user.name;
    const patcher = (user, name) => {user.name = name};
    
    const [name, dispatch] = store.useReducer(myReducer, "user", {selector, patcher});

    // Other stuff
}
```
<br/>

### store.getState
`store.getState` is used to get a global state object from a store, it accepts one required parameter which is a key(string) and another optional parameters which is the configuration object(available configurations are `default` and `persist` works just like in `store.setState`). When called, `store.getState` returns a global state object.

```js
// Signature
store.getState(key: String, config?: {default: Any, persist: Boolean})
```

Here is how to use it

```js
const globalState = store.getState(key);
```

<br/>

### store.subscribe & globalState.subscribe
If you want to listen to changes in a store you can subscribe to it by using `store.subscribe`. it accepts an observer function. For example 

```js
// Subscribe to store changes
const unsubscribe = store.subscribe(function(key: String, value: Any){
    // key is the key for a global state that has changed 
    // value is the new value of a global state
})

// You can unsubscribe by calling the result
unsubscribe();
```

If you want to subscribe to a single global state you can use 

```js
// Subscribe to store changes
const unsubscribe = store.getState(key).subscribe(function(value){
    // value is the new value of a global state
})

// You can unsubscribe by calling the result
unsubscribe();
```

You can even subscribe to a deeply nested state by using a selector as 

```js
store.getState(key).subscribe({
    observer:  function(value){
        // value is the new value of a global state 
    },
    selector: function(value){
        return  selected_state
    })
})
```
With this observer function will only be called when the selected state changes

<br/>

### store.remove
This is used to remove a global state from a store if you don't need it anymore or you want to reload/reset it. It accepts a key of a global state or a list of keys to remove and a function to run after a removal. Note the function runs before components subscribed to removed global state(s) re-renders.

```js
// Signature
store.remove(key: String, fn: Function)
```

Below is an example showing how to use it

```js
import React from 'react';
import { createStore } from 'state-pool';


const store = createStore();
store.setState("count", 0);

function ClicksCounter(props){
    const [count, setCount, updateCount] = store.useState("count");

    const incrementCount = (e) => {
        setCount(count+1);
    }

    const resetCounter = (e) => {
        store.remove("count", initializeStore)
    }

    return (
        <div>
            Count: {count}
            <br/>
            <button onClick={incrementCount}>Click</button>
            <button onClick={resetCounter}>Reset</button>
        </div>
    );
}

ReactDOM.render(ClicksCounter, document.querySelector("#root"));
```

From an example above, when you click `Reset` button `store.remove` will remove `count` global state and create it again by executing `initializeStore`.


**NOTE:** If we had more than one state to delete we could do

```js
store.remove([key1, key2, key3, ...], initializeStore);
```

<br/>

### store.clear
This is used to clear an entire store if you don't need all global states in it anymore or you want to reload/reset all global states. It accepts a function to run after clearing the store. Note the function runs before components subscribed to all global states in a store rerenders.

```js
// Signature
store.clear(fn: Function)
```

Below is an example showing how to use it

```js
import React from 'react';
import { createStore } from 'state-pool';


const store = createStore();

const user = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}
    
const profile = {
    url: "https://yezyilomo.me",
    rating: 5
}
store.setState("user", user);
store.setState("profile", profile);


function UserInfo(props){
    const [user, setUser, updateUser] = store.useState("user");

    const updateName = (e) => {
        updateUser(user => {
            user.name = e.target.value;
        });
    }

    const resetStore = (e) => {
        store.clear(initializeStore);
    }

    return (
        <div>
            Name: {user.name}
            <br/>
            <input type="text" value={user.name} onChange={updateName}/>
             <button onClick={resetStore}>Reset Store</button>
        </div>
    );
}

ReactDOM.render(UserInfo, document.querySelector("#root"));
```
From the code above, when you click `Reset Store` button `store.clear` will remove all global states from the store and create them again by executing `initializeStore`. This might come in handy when you need to clear all data when user logs out of your application.


**NOTE:** both `store.remove` and `store.clear` when executed causes all components subscribed to global states which are removed to rerender.

<br/>

### store.persist(State Persistence)
Sometimes you might want to save your global states in a permanent storage probably because you might not want to lose them when the application is closed(i.e you want to retain them when your application starts).

**State pool** makes it very easy to save your global states in your preferred permanent storage, all you need to do is:

1. Tell state-pool how to save your global state by using `store.persist`
2. Use `persist` configuration to tell state-pool to save your global state in your preferred storage when creating your global state.

When telling **state-pool**  how to save global state to a permanent storage we need to implement four functions which are 

1. `saveState`: This is for saving your global state to your preferred permanent storage, it should accept a `key` as the first parameter, `value` as the second parameter and `isInitialSet` as the third parameter, the third parameter is boolean which tells if the state is being saved for the first time(initial set) or it's just an update. This function is called automatically when `store.setState` is executed and when the  global state changes
2. `loadState`: This is used for loading state from your preferred permanent storage, it should accept a `key` as the only parameter. This function is called when `store.setState` needs an initial value from your storage to populate a global state
3. `removeState`: This is used for removing state from a permanent storage, it should accept a `key` as the only parameter. This function is called when `store.remove` is executed
4. `clear`: This is used for clearing an entire permanent storage, it doesn’t accept any parameter. This function is called when `store.clear` is executed.

Now the way to implement these is by calling `store.persist` and pass them as shown below 

```js
store.persist({
    saveState: function(key, value, isInitialSet){/*your code to save state */},
    loadState: function(key){/*your code to load state */},
    removeState: function(key){/*your code to remove state */},
    clear: function(){/*your code to clear storage */}
})
```

After implementing these four functions you're good to go, you won’t need to worry about calling them, **state-pool** will be doing that for you automatically so that you can focus on using your states.

As discussed earlier both `store.setState`, `store.useState` and `store.useReducer` accepts an optional configuration parameter, `persist` being one of configurations, this is the one which is used to tell **state-pool** whether to save your global state to a permanent storage or not. i.e

```js
store.setState(
    key: String,
    initialState: Any,
    config?: {persist: Boolean}
)
```

```js
store.useState(
    key: String,
    config?: {default: Any, persist: Boolean, ...otherConfigs}
)
```

```js
store.useReducer(
    reducer: Function,
    key: String,
    config?: {default: Any, persist: Boolean, ...otherConfigs}
)
```

By default the value of `persist` in all cases is false(which means it doesn't save global states to a permanent storage), so if you want to activate it, you have to set it to be true.

What's even better about **state-pool** is that you get the freedom to choose what to save in your permanent storage and what's not to, so you don't need to save the whole store in your permanent storage, but if you want to save the whole store you can use `PERSIST_ENTIRE_STORE` configuration.

Below is an example showing how you could implement state persistance in local storage.

```js
import { createStore } from 'state-pool';

const store = createStore();

let timerId: any = null
const DEBOUNCE_TIME = 1000  // In milliseconds

store.persist({
    PERSIST_ENTIRE_STORE: true,  // Use this only if you want to persist the entire store
    saveState: function(key, value, isInitialSet){
        const doStateSaving = () => {
            try {
                const serializedState = JSON.stringify(state);
                window.localStorage.setItem(key, serializedState);
            } catch {
                // Ignore write errors
            }
        }

        if(isInitialSet){
            // We don't debounce saving state since it's the initial set
            // so it's called only once and we need our storage to be updated
            // right away
            doStateSaving();
        }
        else {
            // Here we debounce saving state because it's the update and this function
            // is called every time the store state changes. However, it should not
            // be called too often because it triggers the expensive `JSON.stringify` operation.
            clearTimeout(timerId);
            timerId = setTimeout(() => {
                
            }, DEBOUNCE_TIME);
        }
    },
    loadState: function(key){
        try {
            const serializedState = window.localStorage.getItem(key);
            if (serializedState === null) {
                // No state saved
                return undefined
            }
            return JSON.parse(serializedState);
        } catch (err) {
            // Failed to load state
            return undefined
        }
    },
    removeState: function(key){
        window.localStorage.removeItem(key);
    },
    clear: function(){
        window.localStorage.clear();
    }
})

```

**Note:** When you set `PERSIST_ENTIRE_STORE = true`, **state-pool** will be persisting all your global states to the permanent storage by default unless you explicitly specify `persist = false` when initializing your global state.

<br/>

## Using state pool without a store(Low level core API)
Managing global states by organizing them in a single store works great but not everyone likes it, some prefer not having a single central container for all global states. State pool allows you to create as many stores as you want and use them anywhere in your application, it doesn't enforce having a single store.

**State pool** is not built around the concept of store, it has its own API from which a store is implemented. In **state-pool** a store is nothing but a container for global states, it's more like a variable which hold global states, so you can have as many of these variables as you want.

Since **state-pool**'s core API is not built around a store, this means you can use it to manage global states without using a store. **state-pool** doesn't care where you store your global state as long as you can access them, for-instance I could choose to store my global state in a global variable and it would still work just fine.

Here is an example
```js
// Managing state without using a store
import React from 'react';
import { createGlobalState, useGlobalState } from 'state-pool';


const count = createGlobalState(0);  // Create global state and initialize it with 0

function ClicksCounter(props){
    // Use count global state
    const [count, setCount, updateCount] = useGlobalState(count);

    const incrementCount = (e) => {
        setCount(count+1);
    }

    return (
        <div>
            Count: {count}
            <br/>
            <button onClick={incrementCount}>Click</button>
        </div>
    );
}

ReactDOM.render(ClicksCounter, document.querySelector("#root"));
```

<br/>

Now let's explore the low level core API of **state-pool**

### createGlobalState
This is the basic unit of **state-pool**, it's a function which is used to create a global state object, it accepts one parameter which is the initial value.

```js
// Signature
createGlobalState(initialValue: Any)
```

Here is how to use it

```js
import { createGlobalState } from 'state-pool';

const userName = createGlobalState("Yezy");
```

Some of the methods available in global state object are `getValue`, `updateValue` and `subscribe`

- `getValue`: This is used to get the value of a global state
  ```js
  // Signature
  globalState.getValue(selector?: Function);
  ```
- `updateValue`: This is used to update the value of a global state
  ```js
  // Signature
  globalState.updateValue(updater: Function, config?: {selector, patcher});
  ```
- `subscribe`: This is used to listen to all changes from a global state
  ```js
  // Signature
  globalState.subscribe(observer: Function | Subscription: {observer, selector});
  ```

Below is an example showing all of them in action
```js
import { createGlobalState } from 'state-pool';

const count = createGlobalState(0);

count.getValue()  // This will give 0

count.updateValue(val => val+1)  // This will increment the value of count

// This will print whenever count change
const unsubscribe = count.subscribe(val => console.log(val)) 

unsubscribe()  // This will unsubscribe the observer above
```

**Note:** `createGlobalState` should be used outside of a react component. 

**FYI:** `createGlobalState` is used to implement `store.setState` API.

<br/>

### useGlobalState hook
`useGlobalState` is a hook that used within a react component to subscribe to a global state. `useGlobalState` works just like `React.useState` hook but it accepts a global state object and returns an array of `[state, setState, updateState]` rather than `[state, setState]`. In addition to a global state object parameter it also accept another optional parameter which is the config object, available configurations are `selector` & `patcher`, these parameters works exactly the same as in `store.useState`. We could say `useGlobalState` is a low level implementation of `store.useState`.

```js
// Signature
useGlobalState(globalState: GlobalState, config?: {selector: Function, patcher: Function})
```

Below is a simple example showing how to use `useGlobalState` hook

```js
import React from 'react';
import { createGlobalState, useGlobalState } from 'state-pool';


const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

const user = createGlobalState(initialGlobalState);

function Component(props){
    const [user, setUser, updateUser] = useGlobalState(user);
    // Other stuff ...
}
```

Below is an example with `selector` and `patcher` configurations

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

const user = createGlobalState(initialGlobalState);


function UserName(props){
    const selector = (user) => user.name;  // Subscribe to user.name only
    const patcher = (user, name) => {user.name = name};  // Update user.name

    const [name, setName] = useGlobalState(user, {selector: selector, patcher: patcher});

    const handleNameChange = (e) => {
        setName(e.target.value);
    }

    return (
        <div>
            Name: {name}
            <br/>
            <input type="text" value={name} onChange={handleNameChange}/>
        </div>
    );
}
```

**FYI:** `useGlobalState` is used to implement `store.useState` hook.

<br/>

### useGlobalStateReducer hook
This is an alternative to `useGlobalState`, it works just like `React.useReducer` hook((If you’re familiar with `React.useReducer`, you already know how this works). It accepts a reducer and a global state object as parameters, it returns the current state paired with a dispatch method. In addition to the two parameters it also accept another optinal perameter which is the configuration object, available configurations are `selector` and `patcher` they work exactly the same just like in `useGlobalState`.


```js
// Signature
useGlobalStateReducer(
    reducer: Function,
    globalState: GlobalState,
    config?: {selector: Function, patcher: Function}
)
```

Below is a simple example showing how to use `useGlobalStateReducer`

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

const user = createGlobalState(initialGlobalState);

function myReducer(state, action){
    // This could be any reducer
    // Do whatever you want to do here
    return newState
}

function Component(props){
    const [name, dispatch] = useGlobalStateReducer(myReducer, user);
    
    // Other stuff ...
}
```

Below is an example with `selector` and `patcher` parameters

```js
const initialGlobalState = {
    name: "Yezy",
    age: 25,
    email: "yezy@me.com"
}

const user = createGlobalState(initialGlobalState);


function myReducer(state, action){
    // This could be any reducer
    // Do whatever you want to do here
    return newState
}

function UserInfo(props){
    const selector = (user) => user.name;
    const patcher = (user, name) => {user.name = name};
    
    const [name, dispatch] = useGlobalStateReducer(myReducer, user, {selector, patcher});

    // Other stuffs
}
```

**FYI:** `useGlobalState` hook is derived from `useGlobalStateReducer` hook, also this hook is used to implement `store.useReducer`.

<br/>

## Typing state
All state related functions support implicity and explicity typing 

```ts
store.setState<number>('count', 0);

store.useState<number>('count');

store.useReducer<number>(reducer, 'count');

// For none key based
const count = createGlobalState<number>(0);

useGlobalState<number>(count);

useGlobalStateReducer<number>(reducer, count);


// Typing with selector
store.setState<{name: string, age: number}>('user', {name: 'Yezy', age: 25});

store.useState<string>('user', {selector: user => user.name});
store.useState<number>('age', {selector: user => user.age});

store.useReducer<string>(reducer, 'user', {selector: user => user.name});
store.useReducer<number>(reducer, 'user', {selector: user => user.age});

// For none key based
const user = createGlobalState<{name: string, age: number}>({name: 'Yezy', age: 25});

useGlobalState<string>(user, {selector: user => user.name});
useGlobalState<number>(user, {selector: user => user.age});

useGlobalStateReducer<string>(reducer, user, {selector: user => user.name});
useGlobalStateReducer<number>(reducer, user, {selector: user => user.age});
```

Pretty cool, right?
